;PETCB2 music data player
;only supports the expanded 16-bit divider mode, as the regular CB2 is trivial



PIA1_PORTA=$e810
PIA1_PORTB=$e812



.segment "PETPLAY_ZP" : zeropage

PP_IRQ_SR:			.res 1
PP_IRQ_SR_COPY:		.res 1
PP_PITCH_PREV:		.res 1
PP_PCR_MASK:		.res 1
PP_SRC:				.res 2



.segment "PETPLAY_VARS"

PP_MUSIC_PTR:		.res 2
PP_LOOP_PTR:		.res 2
PP_WAIT:			.res 1
PP_MARKER:			.res 1
PP_WAVE:			.res 1
PP_PITCH:			.res 2



.segment "PETPLAY_CODE"

;-------------------------------------------------------------------------
; initialize player and timer interrupt handler
;-------------------------------------------------------------------------

petplay_init:

	ora #$c0
	sta PP_PCR_MASK

setup_irq_handler:

	sei
	
	jsr pp_get_irq_entry_ptr
	
	ldy #0
	lda (PP_SRC),y
	sta pp_old_irq+1
	iny
	lda (PP_SRC),y
	sta pp_old_irq+2
	lda #>pp_irq_handler
	sta (PP_SRC),y
	lda #<pp_irq_handler
	dey
	sta (PP_SRC),y
		
	lda #$7f			;clear all interrupts
	sta VIA_IER
	lda #$c0			;enable T1 time-out IRQ
	sta VIA_IER
	
	cli

	lda #$ff
	sta <PP_IRQ_SR_COPY
	sta <PP_PITCH_PREV
	
	jmp petplay_stop
	
	
	
;-------------------------------------------------------------------------
; mute the sound and restore timer interrupt handler
;-------------------------------------------------------------------------

petplay_done:

	sei
	
	lda #$7f			;disable all interrupts
	sta VIA_IFR
	sta VIA_IER			;follow with a 4t delay before cli to avoid some 'ghost interrutps'
	
	jsr pp_get_irq_entry_ptr
	
	lda pp_old_irq+1
	ldy #0
	sta (PP_SRC),y
	lda pp_old_irq+2
	iny
	sta (PP_SRC),y
	
	cli
	
	lda #0				;disable shift register to restore datasette access
	sta VIA_CR
	
	jmp pp_set_wave		;jsr:rts
	
	
	
;-------------------------------------------------------------------------
; update sound state, called 50/60 times per second from the vblank interrupt
;-------------------------------------------------------------------------

pp_update:

	lda PP_MUSIC_PTR+0
	ora PP_MUSIC_PTR+1
	bne :+
	rts
	
:

	lda PP_WAIT
	beq @read0
	dec PP_WAIT
	beq @read0
	rts
	
@read0:

	lda PP_MUSIC_PTR+0
	sta PP_SRC+0
	lda PP_MUSIC_PTR+1
	sta PP_SRC+1
	
	ldy #0
	lda (PP_SRC),y
	bne @noloop

@goloop:

	lda PP_LOOP_PTR+0
	sta PP_MUSIC_PTR+0
	lda PP_LOOP_PTR+1
	sta PP_MUSIC_PTR+1
	jmp @read0
	
@noloop:

	cmp #$ff
	bne @read1
	sta PP_MARKER
	inc PP_MUSIC_PTR+0
	bne @read0
	inc PP_MUSIC_PTR+1
	jmp @read0

@read1:

	sta PP_WAIT
	iny
	lda (PP_SRC),y
	sta PP_WAVE
	iny
	tax
	beq @nopitch
	
	lda (PP_SRC),y
	sta PP_PITCH+0
	iny
	lda (PP_SRC),y
	sta PP_PITCH+1
	iny
	
@nopitch:

	tya
	clc
	adc PP_MUSIC_PTR+0
	sta PP_MUSIC_PTR+0
	bcc :+
	inc PP_MUSIC_PTR+1
:

	lda PP_WAVE
	jsr pp_set_wave
	ldx PP_PITCH+0
	ldy PP_PITCH+1
	jmp pp_set_pitch	;jsr:rts
	
	
	
;-------------------------------------------------------------------------
; start playing a song, X:A is an absolute pointer to the music data
;-------------------------------------------------------------------------

petplay_start:

	pha
	jsr petplay_stop
	pla
	
	sta PP_MUSIC_PTR+0
	sta PP_LOOP_PTR+0
	stx PP_MUSIC_PTR+1
	stx PP_LOOP_PTR+1
	
	rts
	
	
	
;-------------------------------------------------------------------------
; stop playing a song
;-------------------------------------------------------------------------

petplay_stop:

	lda #0
	sta PP_MUSIC_PTR+0
	sta PP_MUSIC_PTR+1
	sta PP_WAIT
	sta PP_MARKER
	sta PP_WAVE
	sta PP_PITCH+0
	sta PP_PITCH+1
	
	jmp pp_set_wave		;jsr:rts	
	
	

;-------------------------------------------------------------------------
	
pp_get_irq_entry_ptr:

	ldx #0				;check ROM version to install handler
	lda $91
	and #$f0
	cmp #$e0
	beq :+

	lda #$19			;ROM 1.0
	ldx #$02
	bne :++
:
	lda #$90			;ROM 2.0
	ldx #$00
:
	sta <PP_SRC+0
	stx <PP_SRC+1
	
	rts
	
	
	
pp_irq_handler:

	;no need to store registers, monitor already saved them all

	lda VIA_IFR			
	bpl @standard

@custom:

	lda VIA_T1CL		;4 clear interrupt
	lda <PP_IRQ_SR		;3
	cmp #$80			;2
	rol a				;2
	sta <PP_IRQ_SR		;3
	
	and #$20			;2	bit 5 actually gets outputted, speed optimization
	ora PP_PCR_MASK		;3
	sta VIA_PCR			;4

	pla					;exit early
	tay
	pla
	tax
	pla
	rti
	
@standard:

	lda PIA1_PORTB      ;this acknowledges frame interrupt, otherwise it'll get invoked once again right after cli or rti
	cli					;now interrupts can be re-enabled, so timer1 interrupt can run while standard handler is working,
						;removing distortions from the sound

	jsr pp_update
	
pp_old_irq:
	jmp 0				;self-modified code to avoid jmp (addr) alignment issues
	
	
	
pp_set_wave:

	pha
	cmp <PP_IRQ_SR_COPY
	beq :+
	sta <PP_IRQ_SR_COPY
	sta <PP_IRQ_SR
	sta VIA_SR
:
	pla
	rts
	


;Y:X 16-bit pitch

pp_set_pitch:

	tya
	beq @sr
	
@irq:

	txa
	asl a
	sta VIA_T1CL
	tya
	rol a
	sta VIA_T1CH

	lda #$c0			;enable T1 free running interrupts
	sta VIA_CR
	
	rts
	
@sr:

	lda #$10			;disable T1 free running interrupts, shift out free running at T2 rate
	sta VIA_CR
	stx VIA_T2CL
	
	lda <PP_IRQ_SR_COPY		;reload shift register to start shifting, this issue mentioned in pet-io-2.txt
	sta VIA_SR
	
	rts