#include "petcb2.h"
#include "gui.h"



extern void* hInstance;

inline HINSTANCE GetInstance() { return (HINSTANCE)hInstance; }



const char* GUI::GetClassName(void)
{
	sprintf(ClassName, "PCSPE%08x", (unsigned long)(LONG_PTR)this);

	return (const char*)ClassName;
}



GUI::GUI(AudioEffect* effect)
{
	WNDCLASS wc;

	pcspe=(PETCB2*)effect;

	rect.left  =0;
	rect.top   =0;
	rect.right =GUI_WINDOW_WDT;
	rect.bottom=GUI_WINDOW_HGT;

	wc.style=CS_GLOBALCLASS;
	wc.lpfnWndProc=(WNDPROC)WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=GetInstance();
	wc.hIcon=0;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=GetSysColorBrush(COLOR_BTNFACE);
	wc.lpszMenuName=0;
	wc.lpszClassName=GetClassName(); 

	RegisterClass(&wc);
}



GUI::~GUI()
{
	UnregisterClass(GetClassName(),GetInstance());
}
	
	
	
bool GUI::open(void* ptr)
{
	char tempstr[32];

	AEffEditor::open(ptr);

	HWND parent=HWND(ptr);
	HINSTANCE hinst=GetInstance();

	hWndForm = CreateWindow(GetClassName(), "PCSPE", WS_CHILD | WS_VISIBLE, 0, 0, (rect.right - rect.left), (rect.bottom - rect.top), parent, 0, hinst, this);

	SetWindowLongPtr(hWndForm,GWLP_USERDATA,(LONG_PTR)this);

	VstInt32 style=WS_CHILD|WS_VISIBLE;

	VstInt32 x=10;
	VstInt32 y=14;
	VstInt32 w=60;
	VstInt32 h=25;

	hWndLabelC = CreateWindow("static", "ST_U", style, x, y, 100, h, hWndForm, NULL, hinst, NULL); y += 40;
	hWndLabelV = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndLabelA = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndLabelP = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndLabelW = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 40;
	hWndLabelR = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	SetWindowText(hWndLabelC, "MIDI channel:");
	SetWindowText(hWndLabelV, "Volume:");
	SetWindowText(hWndLabelA, "Arp:");
	SetWindowText(hWndLabelP, "Pitch:");
	SetWindowText(hWndLabelW, "Wave:");
	SetWindowText(hWndLabelR, "Range:");

	x=110;
	y=10;
	w=24;
	h=24;

	style=WS_VISIBLE|WS_CHILD|BS_DEFPUSHBUTTON|BS_OWNERDRAW|BS_NOTIFY;

	for(VstInt32 i=0;i<16;++i)
	{
		sprintf(tempstr,"%i",i+1);

		hWndChannel[i]=CreateWindow("button",tempstr,style,x,y,w,h,hWndForm,NULL,hinst,NULL);

		x+=w+6;
	}

	style=WS_CHILD|WS_VISIBLE|ES_LEFT|WS_BORDER|ES_AUTOHSCROLL;

	x=80;
	y=50;
	w=GUI_WINDOW_WDT-10-x;
	h=25;

	hWndEditV = CreateWindow("edit", "0", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndEditA = CreateWindow("edit", "0", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndEditP = CreateWindow("edit", "0", style, x, y, w, h, hWndForm, NULL, hinst, NULL); y += 30;
	hWndEditW = CreateWindow("edit", "0", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	SendMessage(hWndEditV, EM_SETLIMITTEXT, ENVELOPE_TEXT_LEN, 0);
	SendMessage(hWndEditA, EM_SETLIMITTEXT, ENVELOPE_TEXT_LEN, 0);
	SendMessage(hWndEditP, EM_SETLIMITTEXT, ENVELOPE_TEXT_LEN, 0);
	SendMessage(hWndEditW, EM_SETLIMITTEXT, ENVELOPE_TEXT_LEN, 0);

	SetWindowLongPtr(hWndEditV, GWLP_USERDATA, 0xdeadf00b);	//Reaper hack to prevent hijacking Space button
	SetWindowLongPtr(hWndEditA, GWLP_USERDATA, 0xdeadf00b);
	SetWindowLongPtr(hWndEditP, GWLP_USERDATA, 0xdeadf00b);
	SetWindowLongPtr(hWndEditW, GWLP_USERDATA, 0xdeadf00b);

	style = WS_VISIBLE | WS_CHILD | BS_DEFPUSHBUTTON | BS_OWNERDRAW | BS_NOTIFY;

	y = 180;

	h = 24;
	w = 80;
	x = 80;

	hWndRangeStandard = CreateWindow("button", "Standard", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	x += w + 14;

	hWndRangeExpanded = CreateWindow("button", "Expanded", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	h = 24;
	w = 24;
	x = GUI_WINDOW_WDT - w - 14;

	hWndLogFileAuto = CreateWindow("button", "AC", style, x, y, w, h, hWndForm, NULL, hinst, NULL); 
	
	w = 70;
	x -= w + 14;
	
	hWndLogFile = CreateWindow("button", "Log start", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	w = 50;
	x -= w + 14;
	hWndMultiMode = CreateWindow("button", "Multi", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	SetUpdate(true);

	return true;
}



void GUI::close()
{
	DestroyWindow(hWndLabelC);
	DestroyWindow(hWndLabelV);
	DestroyWindow(hWndLabelA);
	DestroyWindow(hWndLabelP);
	DestroyWindow(hWndLabelW);
	DestroyWindow(hWndEditV);
	DestroyWindow(hWndEditA);
	DestroyWindow(hWndEditP);
	DestroyWindow(hWndEditW);

	for (VstInt32 i = 0; i < 16; ++i) DestroyWindow(hWndChannel[i]);

	DestroyWindow(hWndRangeStandard);
	DestroyWindow(hWndRangeExpanded);
	DestroyWindow(hWndLogFile);
	DestroyWindow(hWndLogFileAuto);
	DestroyWindow(hWndMultiMode);

	DestroyWindow(hWndForm);

	AEffEditor::close();
}



bool GUI::getRect(ERect **ppErect)
{
   *ppErect=&rect;

   return true;
}



void GUI::idle(void)
{
	char label[1024], newlabel[1024];

	VstInt32 pgm = pcspe->MidiProgram[pcspe->MidiChannel];

	if (NeedUpdateEdits)
	{
		NeedUpdateEdits = false;

		SetWindowText(hWndEditV, pcspe->EnvelopeV[pgm]);
		SetWindowText(hWndEditA, pcspe->EnvelopeA[pgm]);
		SetWindowText(hWndEditP, pcspe->EnvelopeP[pgm]);
		SetWindowText(hWndEditW, pcspe->EnvelopeW[pgm]);
	}

	if (NeedUpdate)
	{
		NeedUpdate = false;

		GetWindowText(hWndLogFile, label, GetWindowTextLength(hWndLogFile) + 1);

		strcpy(newlabel, pcspe->LogActive ? "Log stop" : "Log start");

		if (strcmp(label, newlabel))
		{
			SetWindowText(hWndLogFile, newlabel);

			InvalidateRect(hWndLogFile, NULL, true);
		}

		InvalidateRect(hWndRangeStandard, NULL, true);
		InvalidateRect(hWndRangeExpanded, NULL, true);
		InvalidateRect(hWndMultiMode, NULL, true);
		InvalidateRect(hWndLogFileAuto, NULL, true);
		InvalidateRect(hWndLabelV, NULL, true);
		InvalidateRect(hWndLabelA, NULL, true);
		InvalidateRect(hWndLabelP, NULL, true);
		InvalidateRect(hWndLabelW, NULL, true);

		for (VstInt32 i = 0; i < 16; ++i) InvalidateRect(hWndChannel[i], NULL, true);

		pcspe->updateDisplay();
	}
}



void GUI::SetUpdate(bool program)
{
	NeedUpdate = true;
	NeedUpdateEdits = program;
}



LRESULT WINAPI GUI::WndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	char tempstr[16],ext[1024];
	bool compile,highlight,noext;

	GUI* gui = (GUI*)GetWindowLongPtr(hWnd, GWLP_USERDATA);

	HWND phWnd = (HWND)lParam;
	HDC phDC = (HDC)wParam;

	switch(message)
	{
	case WM_COMMAND:
		{
			if(HIWORD(wParam)==EN_CHANGE)	//any edit has been changed
			{
				compile=false;

				if (phWnd == gui->hWndEditV) { GetWindowText(phWnd, gui->pcspe->EnvelopeV[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]], ENVELOPE_TEXT_LEN); compile = true; }
				if (phWnd == gui->hWndEditA) { GetWindowText(phWnd, gui->pcspe->EnvelopeA[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]], ENVELOPE_TEXT_LEN); compile = true; }
				if (phWnd == gui->hWndEditP) { GetWindowText(phWnd, gui->pcspe->EnvelopeP[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]], ENVELOPE_TEXT_LEN); compile = true; }
				if (phWnd == gui->hWndEditW) { GetWindowText(phWnd, gui->pcspe->EnvelopeW[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]], ENVELOPE_TEXT_LEN); compile = true; }

				if(compile)
				{
					gui->pcspe->CompileInstruments();

					gui->SetUpdate(false);
				}

				return 0;
			}

			for (VstInt32 i = 0; i < 16; ++i)
			{
				if(phWnd==gui->hWndChannel[i])
				{
					if(gui->pcspe->MidiChannel!=i)
					{
						gui->pcspe->MidiChannel=i;

						gui->SetUpdate(true);
					}

					return 0;
				}
			}

			if (phWnd == gui->hWndRangeStandard)
			{
				if (HIWORD(wParam) == BN_CLICKED)
				{
					gui->pcspe->RangeExpanded = false;

					gui->SetUpdate(false);
				}

				return 0;
			}

			if (phWnd == gui->hWndRangeExpanded)
			{
				if (HIWORD(wParam) == BN_CLICKED)
				{
					gui->pcspe->RangeExpanded = true;

					gui->SetUpdate(false);
				}

				return 0;
			}

			if(phWnd==gui->hWndLogFileAuto)
			{
				if(HIWORD(wParam)==BN_CLICKED)
				{
					gui->pcspe->LogAuto^=true;

					gui->SetUpdate(false);
				}

				return 0;
			}

			if(phWnd==gui->hWndMultiMode)
			{
				if(HIWORD(wParam)==BN_CLICKED)
				{
					gui->pcspe->MultiMode[gui->pcspe->MidiChannel]^=true;

					gui->SetUpdate(false);
				}

				return 0;
			}

			if(phWnd==gui->hWndLogFile)
			{
				if(HIWORD(wParam)==BN_CLICKED)
				{
					if(gui->pcspe->LogActive)
					{
						gui->pcspe->LogStop();
					}
					else
					{
						OPENFILENAME ofn;
						ZeroMemory(&ofn,sizeof(ofn));
						ofn.lStructSize=sizeof(ofn);

						ofn.hwndOwner=hWnd;
						ofn.nFilterIndex=1;
						ofn.nMaxFileTitle=0;
						ofn.Flags=OFN_ENABLESIZING|OFN_OVERWRITEPROMPT;

						ofn.lpstrFileTitle=NULL;
						ofn.lpstrFilter="ASM file\0*.s\0C header\0*.h\0Binary file\0*.bin\0All\0*.*\0";
						ofn.lpstrInitialDir=NULL;
						ofn.lpstrFile=gui->pcspe->LogFileName;
						ofn.nMaxFile=MAX_PATH;
						ofn.lpstrTitle="Save VIA log file";
						ofn.Flags = OFN_OVERWRITEPROMPT;

						if(GetSaveFileName(&ofn))
						{
							strcpy(gui->pcspe->LogFileName,ofn.lpstrFile);

							strcpy(ext,ofn.lpstrFile);

							noext=true;

							for (VstInt32 i = (VstInt32)strlen(ext) - 1; i >= 0; --i)
							{
								if(ext[i]=='.')
								{
									strcpy(ext,&ext[i+1]);
									noext=false;
									break;
								}
							}

							if(noext)
							{
								switch(ofn.nFilterIndex)
								{
								case 1: strcat(gui->pcspe->LogFileName,".s"); strcpy(ext,"s"); break;
								case 2: strcat(gui->pcspe->LogFileName,".h"  ); strcpy(ext,"h"  ); break;
								case 3: strcat(gui->pcspe->LogFileName,".bin"); strcpy(ext,"bin"); break;
								}
							}

							gui->pcspe->LogFormat = LOG_FORMAT_H;

							if (!strcmp(ext, "s") || !strcmp(ext, "S")) gui->pcspe->LogFormat = LOG_FORMAT_ASM;
							if (!strcmp(ext, "asm") || !strcmp(ext, "ASM")) gui->pcspe->LogFormat = LOG_FORMAT_ASM;
							if (!strcmp(ext, "h") || !strcmp(ext, "H")) gui->pcspe->LogFormat = LOG_FORMAT_H;
							if (!strcmp(ext, "bin") || !strcmp(ext, "BIN")) gui->pcspe->LogFormat = LOG_FORMAT_BIN;

							gui->pcspe->LogStart();
						}
					}

					return 0;
				}
			}
		}
		break;

	case WM_KEYDOWN:
		{
		}
		return 0;

	case WM_DRAWITEM:
		{
			LPDRAWITEMSTRUCT lpDIS=(LPDRAWITEMSTRUCT)lParam;
		
			highlight=false;

			if (lpDIS->hwndItem == gui->hWndChannel[gui->pcspe->MidiChannel]) highlight = true;

			if (lpDIS->hwndItem == gui->hWndLogFileAuto&&gui->pcspe->LogAuto) highlight = true;

			if (lpDIS->hwndItem == gui->hWndRangeStandard && !gui->pcspe->RangeExpanded) highlight = true;

			if (lpDIS->hwndItem == gui->hWndRangeExpanded && gui->pcspe->RangeExpanded) highlight = true;

			if(lpDIS->hwndItem==gui->hWndMultiMode&&gui->pcspe->MultiMode[gui->pcspe->MidiChannel]) highlight=true;

			VstInt32 color = GetSysColor(highlight ? COLOR_BTNSHADOW : COLOR_BTNFACE);

			SetDCBrushColor(lpDIS->hDC,color);
			SetBkColor(lpDIS->hDC,color);

			SelectObject(lpDIS->hDC,GetStockObject(DC_BRUSH));

			RoundRect(lpDIS->hDC,lpDIS->rcItem.left,lpDIS->rcItem.top,lpDIS->rcItem.right,lpDIS->rcItem.bottom,5,5);

			VstInt32 len = GetWindowTextLength(lpDIS->hwndItem);

			GetWindowTextA(lpDIS->hwndItem,tempstr,len+1);

			lpDIS->rcItem.top+=4;

			DrawTextA(lpDIS->hDC,tempstr,len,&lpDIS->rcItem,DT_CENTER);
		}
        return TRUE;

	case WM_CTLCOLORSTATIC:
		{
			VstInt32 color = GetSysColor(COLOR_BTNFACE);

			if(phWnd==gui->hWndLabelV)
			{
				if(gui->pcspe->Instrument[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]].EnvVol.error)
				{
					color=GUI_COLOR_ERROR;
				}
			}

			if(phWnd==gui->hWndLabelA)
			{
				if(gui->pcspe->Instrument[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]].EnvArp.error)
				{
					color=GUI_COLOR_ERROR;
				}
			}

			if(phWnd==gui->hWndLabelP)
			{
				if(gui->pcspe->Instrument[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]].EnvPitch.error)
				{
					color=GUI_COLOR_ERROR;
				}
			}

			if (phWnd == gui->hWndLabelW)
			{
				if (gui->pcspe->Instrument[gui->pcspe->MidiProgram[gui->pcspe->MidiChannel]].EnvWave.error)
				{
					color = GUI_COLOR_ERROR;
				}
			}

			SetBkColor(phDC,color);
		}
		return (LRESULT)GetSysColorBrush(COLOR_BTNFACE);
	}

	return DefWindowProc(hWnd,message,wParam,lParam);
}