#include "petcb2.h"
#include "gui.h"



AudioEffect* createEffectInstance(audioMasterCallback audioMaster)
{
	return new PETCB2(audioMaster);
}



PETCB2::PETCB2(audioMasterCallback audioMaster) : AudioEffectX(audioMaster, NUM_PROGRAMS, NUM_PARAMS)
{
	setEditor((AEffEditor*)new GUI(this));

	setNumInputs(NUM_INPUTS);
	setNumOutputs(NUM_OUTPUTS);
	setUniqueID(PLUGIN_UID);

	isSynth();
	canProcessReplacing();
	programsAreChunks(true);

	MidiChannel=0;

	for (VstInt32 chn = 0; chn < 16; ++chn)
	{
		MidiProgram[chn] = 0;
		MidiProgramPrev[chn] = 0;
		MultiMode[chn] = false;

		MidiPitchBend[chn] = 0;
		MidiPitchBendRange[chn] = 2.0f;
	}

	for (VstInt32 pgm = 0; pgm < NUM_PROGRAMS; ++pgm)
	{
		sprintf(ProgramName[pgm], "%3.3i default", pgm + 1);

		memset(&EnvelopeV[pgm], 0, ENVELOPE_TEXT_LEN);
		memset(&EnvelopeA[pgm], 0, ENVELOPE_TEXT_LEN);
		memset(&EnvelopeP[pgm], 0, ENVELOPE_TEXT_LEN);
		memset(&EnvelopeW[pgm], 0, ENVELOPE_TEXT_LEN);

		strcpy(EnvelopeV[pgm], "2");
		strcpy(EnvelopeA[pgm], "0");
		strcpy(EnvelopeP[pgm], "0");
		strcpy(EnvelopeW[pgm], "15");
	}

	VstInt32 pgm = 0;
	VstInt32 row = 0;

	while(1)
	{
		if(!strlen(PresetData[row])) break;

		sprintf(ProgramName[pgm],"%3.3i %s",pgm+1,PresetData[row++]);

		strcpy(EnvelopeV[pgm], PresetData[row++]);
		strcpy(EnvelopeA[pgm], PresetData[row++]);
		strcpy(EnvelopeP[pgm], PresetData[row++]);
		strcpy(EnvelopeW[pgm], PresetData[row++]);

		++pgm;
	}

	CompileInstruments();

	SetUpdateGUI();

	for (VstInt32 chn = 0; chn < SYNTH_CHANNELS; ++chn)
	{
		SynthChannel[chn].note=-1;

		SynthChannelOrder[chn]=chn;
	}

	FrameRate=60;

	pOutputGain=.5f;

	RangeExpanded = true;
	LogActive = false;
	LogAuto = false;
	LogAutoEnabled = false;
	strcpy(LogFileName,"");

	sFrameAcc=0;

	sVIAOutput = 0;
	sVIACount = 0;
	sVIAReload = 1193180 / 440;
	sVIAWave = 0;
	sVIAWavePrev = 0;
	sVIAWaveShift = 0;
	sVIAResampleAcc = 0;

	MidiQueue.clear();

	MidiRPNLSB=0;
	MidiRPNMSB=0;
	MidiDataLSB=0;
	MidiDataMSB=0;

	memset(MidiKeyState,0,sizeof(MidiKeyState));

	suspend();
}



PETCB2::~PETCB2()
{
	LogStop();
}



void PETCB2::setParameter(VstInt32 index,float value)
{
}



float PETCB2::getParameter(VstInt32 index)
{
	return 0;
} 



void PETCB2::getParameterName(VstInt32 index,char *label)
{
} 



void PETCB2::getParameterDisplay(VstInt32 index,char *text)
{
} 



void PETCB2::getParameterLabel(VstInt32 index,char *label)
{
} 



VstInt32 PETCB2::getProgram(void)
{
	UpdateGUI();	//force GUI update if current program has been changed while processReplacing is not called (plugin is added but not connected in Psycle)

	return MidiProgram[MidiChannel];
}



void PETCB2::setProgram(VstInt32 program)
{
	if (program > NUM_PROGRAMS - 1) program = NUM_PROGRAMS - 1;

	MidiProgram[MidiChannel] = program;
}



void PETCB2::getProgramName(char* name)
{
	strcpy(name,ProgramName[MidiProgram[MidiChannel]]); 
}



void PETCB2::setProgramName(char* name)
{
	strncpy(ProgramName[MidiProgram[MidiChannel]],name,MAX_NAME_LEN);

	ProgramName[MidiProgram[MidiChannel]][MAX_NAME_LEN-1]='\0';
} 



VstInt32 PETCB2::canDo(char* text)
{
	if(!strcmp(text,"receiveVstEvents"   )) return 1;
	if(!strcmp(text,"receiveVstMidiEvent")) return 1;

	return -1;
}



VstInt32 PETCB2::getNumMidiInputChannels(void)
{
	return 16;
}



VstInt32 PETCB2::getNumMidiOutputChannels(void)
{
	return 0;//no MIDI output
}



VstInt32 PETCB2::SaveStringChunk(char *str,float *dst)
{
	VstInt32 i;

	for(i=0;i<(VstInt32)strlen(str);++i) *dst++=str[i];

	*dst++=0;

	return (VstInt32)strlen(str)+1;
}



VstInt32 PETCB2::LoadStringChunk(char *str,int max_length,float *src)
{
	unsigned int c,ptr;

	ptr=0;

	while(1)
	{
		c=(unsigned int)*src++;
		
		str[ptr++]=c;

		if(!c) break;

		if(ptr==max_length-1)
		{
			str[ptr]=0;
			break;
		}
	}

	return ptr;
}



VstInt32 PETCB2::SavePresetChunk(float *chunk)
{
	int chn,ptr,pgm;

	ptr=0;

	chunk[ptr++]=DATA;

	chunk[ptr++]=MPRG; for(chn=0;chn<16;++chn) chunk[ptr++]=(float)MidiProgram[chn];

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		chunk[ptr++]=PROG; chunk[ptr++]=(float)pgm;

		chunk[ptr++]=NAME; ptr+=SaveStringChunk(ProgramName[pgm],&chunk[ptr]);

		chunk[ptr++] = EVOL; ptr += SaveStringChunk(EnvelopeV[pgm], &chunk[ptr]);
		chunk[ptr++] = EARP; ptr += SaveStringChunk(EnvelopeA[pgm], &chunk[ptr]);
		chunk[ptr++] = EPTC; ptr += SaveStringChunk(EnvelopeP[pgm], &chunk[ptr]);
		chunk[ptr++] = EWAV; ptr += SaveStringChunk(EnvelopeW[pgm], &chunk[ptr]);
	}

	chunk[ptr++]=FRMR; chunk[ptr++]=FrameRate;

	chunk[ptr++]=MMOD; for(chn=0;chn<16;++chn) chunk[ptr++]=MultiMode[chn]?1.0f:0;

	chunk[ptr++]=GAIN; chunk[ptr++]=pOutputGain;
	chunk[ptr++]=DONE;

	return ptr*sizeof(float);//size in bytes
}



void PETCB2::LoadPresetChunk(float *chunk)
{
	int chn,pgm;
	float tag;

	while(1)
	{
		tag=*chunk++;

		if(tag==DONE) break;

		if(tag==MPRG) for(chn=0;chn<16;++chn) MidiProgram[chn]=(int)*chunk++;

		if(tag==PROG) pgm=(int)*chunk++;

		if(tag==NAME) chunk+=LoadStringChunk(ProgramName[pgm],MAX_NAME_LEN,chunk);

		if (tag == EVOL) chunk += LoadStringChunk(EnvelopeV[pgm], ENVELOPE_TEXT_LEN, chunk);
		if (tag == EARP) chunk += LoadStringChunk(EnvelopeA[pgm], ENVELOPE_TEXT_LEN, chunk);
		if (tag == EPTC) chunk += LoadStringChunk(EnvelopeP[pgm], ENVELOPE_TEXT_LEN, chunk);
		if (tag == EWAV) chunk += LoadStringChunk(EnvelopeW[pgm], ENVELOPE_TEXT_LEN, chunk);

		if(tag==FRMR) FrameRate=*chunk++;

		if(tag==MMOD) for(chn=0;chn<16;++chn) MultiMode[chn]=*chunk++?true:false;

		if(tag==GAIN) pOutputGain=*chunk++;
	}

	CompileInstruments();
}



VstInt32 PETCB2::getChunk(void** data,bool isPreset)
{
	int size;

	size=SavePresetChunk(Chunk);

	*data=Chunk;

	return size;
}



VstInt32 PETCB2::setChunk(void* data,VstInt32 byteSize,bool isPreset)
{
	if(byteSize>sizeof(Chunk)) return 0;

	memcpy(Chunk,data,byteSize);

	LoadPresetChunk(Chunk);

	SetUpdateGUI();

	return 0;
}



void PETCB2::SetUpdateGUI(void)
{
	UpdateGuiFlag=true;
}



void PETCB2::UpdateGUI(void)
{
	VstInt32 chn;

	for (chn = 0; chn < MIDI_CHANNELS_ALL; ++chn)	//this is needed because setProgram repeatedly called by the host to obtain all program names
	{
		if (MidiProgramPrev[chn] != MidiProgram[chn])
		{
			MidiProgramPrev[chn] = MidiProgram[chn];

			SetUpdateGUI();
		}
	}

	if (UpdateGuiFlag)
	{
		UpdateGuiFlag = false;

		if (editor) ((GUI*)editor)->SetUpdate(true);

		updateDisplay();
	}
}



const char* PETCB2::CompileEnvelope(char *str,envelopeStruct *env,bool hold)
{
	VstInt32 ptr,len,loop,repeat_len,repeat_cnt;
	VstInt32 c,num,num_prev,prefix,time_scale;
	char *err;

	env->len=0;
	env->loop_ptr=0;
	env->has_loop=false;
	env->error=true;	//error by default, cleared on successful parsing

	err=NULL;

	len=(VstInt32)strlen(str);
	ptr=0;

	loop=-1;			//loop not defined yet
	num_prev=0;			//number to repeat
	time_scale=1;
	repeat_cnt=0;
	repeat_len=0;

	while(ptr<len)
	{
		while(str[ptr]==' ') ++ptr;	//skip spaces

		c=str[ptr];

		if(!c) break;

		if(c>='a'&&c<='z') c-=32;	//case insensitive

		prefix=PARSE_PREFIX_NONE;

		if(c=='-') prefix=PARSE_PREFIX_NEGATIVE;
		if(c=='*') prefix=PARSE_PREFIX_REPEAT;
		if(c=='X') prefix=PARSE_PREFIX_TIME_SCALE;
		if(c=='#') prefix=PARSE_PREFIX_ABSOLUTE;

		if(c=='|'||c=='L')	//loop marker
		{
			if(loop<0)
			{
				loop=env->len;
			}
			else
			{
				err="Loop point redefinition";
				break;
			}

			++ptr;
		}
		else
		if(prefix>PARSE_PREFIX_NONE||(c>='0'&&c<='9'))	//parse number with an optional prefix
		{
			if(prefix>PARSE_PREFIX_NONE)
			{
				++ptr;

				c=str[ptr];

				if(!(c>='0'&&c<='9'))
				{
					err="Prefix is not followed by a number";
					break;
				}
			}

			num=0;

			while(ptr<len)
			{
				c=str[ptr];

				if(c>='a'&&c<='z') c-=32;

				if(c<=32||c=='|'||c=='L'||c=='*') break;

				++ptr;

				if(c>='0'&&c<='9')
				{
					num=num*10+(c-'0');
					continue;
				}

				err="Unexpected character while parsing a number";
				break;
			}

			if(prefix==PARSE_PREFIX_TIME_SCALE)
			{
				time_scale=num;

				if(time_scale<1) time_scale=1;

				continue;
			}

			if(prefix==PARSE_PREFIX_NEGATIVE) num=-num;

			if(prefix==PARSE_PREFIX_REPEAT)
			{
				++repeat_cnt;

				repeat_len=num-1;
		
				num=num_prev;
			}
			else
			{
				repeat_cnt=0;

				num_prev=num;
			}

			if(repeat_cnt>1)
			{
				err="More than one repeat";
				break;
			}

			if(repeat_len<2) repeat_len=1;

			repeat_len*=time_scale;

			if(num<-ENVELOPE_NUM_RANGE) num=-ENVELOPE_NUM_RANGE;
			if(num> ENVELOPE_NUM_RANGE) num= ENVELOPE_NUM_RANGE;

			while(repeat_len--)
			{
				if(prefix==PARSE_PREFIX_ABSOLUTE) env->abs[env->len]=1;

				env->value[env->len++]=num;
			}

			if(env->len>ENVELOPE_DATA_LEN)
			{
				err="Compiled envelope data is too long";
				break;
			}
		}
		else
		{
			err="Unexpected character while parsing string";
			break;
		}
	}

	if(!err&&env->len)	//compiled successfully
	{
		env->error=false;

		if(loop<0)
		{
			if(!hold)
			{
				env->value[env->len++]=0;	//needed for pitch envelope to prevent accumulation without loop

				if(env->len>ENVELOPE_DATA_LEN)
				{
					err="Compiled envelope data is too long";
				}
			}

			env->loop_ptr=env->len-1;	//loop to last position, but loop flag is not set
		}
		else
		{
			env->has_loop=true;
			env->loop_ptr=loop;
		}

		if(env->loop_ptr>env->len-1) env->loop_ptr=env->len-1;
		if(env->loop_ptr<0) env->loop_ptr=0;
	}

	if(err||!env->len)	//make default envelope if error or empty line
	{
		env->len=1;
		env->loop_ptr=0;
		env->value[0]=0;
	}

	return err;
}



void PETCB2::CompileInstruments(void)
{
	for (VstInt32 pgm = 0; pgm < NUM_PROGRAMS; ++pgm)
	{
		CompileEnvelope(EnvelopeV[pgm], &Instrument[pgm].EnvVol, true);
		CompileEnvelope(EnvelopeA[pgm], &Instrument[pgm].EnvArp, true);
		CompileEnvelope(EnvelopeP[pgm], &Instrument[pgm].EnvPitch, false);
		CompileEnvelope(EnvelopeW[pgm], &Instrument[pgm].EnvWave, true);
	}
}



void PETCB2::LogFileOpen(void)
{
	char FileName[MAX_PATH] = { 0 };
	char FileExt[MAX_PATH] = { 0 };

	strncpy(FileName, LogFileName, sizeof(FileName));

	for (VstInt32 i = (VstInt32)strlen(FileName) - 1; i >= 0; --i)
	{
		if (FileName[i] == '.')
		{
			FileName[i] = 0;
			strncpy(FileExt, &FileName[i + 1], sizeof(FileExt));
		}
		if (FileName[i] == '\\' || FileName[i] == '/') { ++i; break; }
	}

	if (LogFilePart > 0)
	{
		char pstr[16];

		snprintf(pstr, sizeof(pstr), "_%i", LogFilePart);

		strncat(FileName, pstr, sizeof(FileName));
	}
	
	strncat(FileName, ".", sizeof(FileName));
	strncat(FileName, FileExt, sizeof(FileName));

	switch (LogFormat)
	{
	case LOG_FORMAT_ASM:
	{
		LogFile = fopen(FileName, "wt");

		fprintf(LogFile, ";%s\n", RangeExpanded ? "16-bit range" : "8-bit range");
	}
	break;

	case LOG_FORMAT_H:
	{
		LogFile = fopen(FileName, "wt");

		char name[1024];

		strncpy(name, LogFileName, sizeof(name));

		VstInt32 i = 0;

		for (i = (VstInt32)strlen(name) - 1; i >= 0; --i)
		{
			if (name[i] == '.') name[i] = 0;
			if (name[i] == '\\' || name[i] == '/') { ++i; break; }
		}

		strncpy(name, &name[i], sizeof(name));

		fprintf(LogFile, "const unsigned char %s_data[]={\n", name);

		fprintf(LogFile, "//%s\n", RangeExpanded ? "16-bit range" : "8-bit range");
	}
	break;

	case LOG_FORMAT_BIN:
	{
		LogFile = fopen(FileName, "wb");
	}
	break;

	default:
		return;
	}
}



void PETCB2::LogFileClose(void)
{
	LogFrameFlush();

	switch (LogFormat)
	{
	case LOG_FORMAT_ASM:
	{
		fprintf(LogFile, "\t.byte $00\n");
	}
	break;

	case LOG_FORMAT_H:
	{
		fprintf(LogFile, "0x00\n};\n");
	}
	break;

	case LOG_FORMAT_BIN:
	{
		fputc(0, LogFile);
	}
	break;
	}

	fclose(LogFile);
}



void PETCB2::LogStart(void)
{
	if(LogActive) return;

	LogFilePart = 0;
	LogFrameDelta = 0;
	LogDiv = 0;
	LogDivPrev = -1;
	LogWave = 0;
	LogWavePrev = -1;
	LogActive = true;
	LogAutoEnabled = false;
	LogMarker = false;

	LogFileOpen();

	SetUpdateGUI();
}



void PETCB2::LogStop(void)
{
	if(!LogActive) return;

	LogFileClose();

	LogActive=false;
	LogAutoEnabled=false;

	SetUpdateGUI();
}



void PETCB2::LogFrameFlush(void)
{
	while (LogFrameDelta > 0 && LogDivPrev >= 0)
	{
		int t = LogFrameDelta;
		if (t > LOG_MAX_FRAMES) t = LOG_MAX_FRAMES;

		switch (LogFormat)
		{
		case LOG_FORMAT_ASM:
		{
			fprintf(LogFile, "\t.byte $%2.2x,$%2.2x", t, LogWavePrev);

			if (LogWavePrev)
			{
				if (!RangeExpanded)
				{
					fprintf(LogFile, ",$%2.2x", LogDivPrev);
				}
				else
				{
					fprintf(LogFile, ",$%2.2x,$%2.2x", LogDivPrev & 255, LogDivPrev / 256);
				}
			}

			fprintf(LogFile, "\n");

			if (LogMarker) fprintf(LogFile, "\t.byte $%2.2x\n", LOG_MARKER_TAG);
		}
		break;

		case LOG_FORMAT_H:
		{
			fprintf(LogFile, "0x%2.2x,0x%2.2x,", t, LogWavePrev);

			if (LogWavePrev)
			{
				if (!RangeExpanded)
				{
					fprintf(LogFile, "0x%2.2x,", LogDivPrev);
				}
				else
				{
					fprintf(LogFile, "0x%2.2x,0x%2.2x,", LogDivPrev & 255, LogDivPrev / 256);
				}
			}

			fprintf(LogFile, "\n");

			if (LogMarker) fprintf(LogFile, "0x%2.2x,\n", LOG_MARKER_TAG);
		}
		break;

		case LOG_FORMAT_BIN:
		{
			fputc(t, LogFile);
			fputc(LogWavePrev, LogFile);

			if (LogWavePrev)
			{
				if (!RangeExpanded)
				{
					fputc(LogDivPrev & 255, LogFile);
				}
				else
				{
					fputc(LogDivPrev & 255, LogFile);
					fputc(LogDivPrev >> 8, LogFile);
				}
			}

			if (LogMarker) fputc(LOG_MARKER_TAG, LogFile);
		}
		break;
		}

		LogMarker=false;

		LogFrameDelta -= LOG_MAX_FRAMES;
	}
}



void PETCB2::LogFrame(void)
{
	if (!LogActive) return;

	if (LogAuto && !LogAutoEnabled) return;

	LogDiv = sVIAReload;
	LogWave = sVIAWave;

	if (LogDivPrev != LogDiv || LogWavePrev != LogWave || LogMarker)
	{
		LogFrameFlush();

		LogDivPrev = LogDiv;
		LogWavePrev = LogWave;
		LogFrameDelta = 0;
	}

	++LogFrameDelta;
}



void PETCB2::MidiAddNote(VstInt32 delta,VstInt32 channel,VstInt32 note,VstInt32 velocity)
{
	MidiQueueStruct entry;

	entry.type    =mEventTypeNote;
	entry.channel =channel;
	entry.delta   =delta;
	entry.note    =note;
	entry.velocity=velocity;//0 for key off

	MidiQueue.push_back(entry);
}



void PETCB2::MidiAddPitchBend(VstInt32 delta,VstInt32 channel,float depth)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypePitchBend;
	entry.channel=channel;
	entry.delta  =delta;
	entry.depth  =depth;

	MidiQueue.push_back(entry);
}



bool PETCB2::MidiIsAnyKeyDown(void)
{
	int i;

	for(i=0;i<128;++i) if(MidiKeyState[i]) return true;

	return false;
}



void PETCB2::MidiAddProgramChange(VstInt32 delta,VstInt32 channel,VstInt32 program)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeProgram;
	entry.channel=channel;
	entry.delta  =delta;
	entry.program=program;

	MidiQueue.push_back(entry);
}



void PETCB2::MidiAddControl(VstInt32 delta,VstInt32 channel,VstInt32 value)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeControl;
	entry.channel=channel;
	entry.delta  =delta;
	entry.value  =value;

	MidiQueue.push_back(entry);
}



VstInt32 PETCB2::processEvents(VstEvents* ev) 
{ 
	for (VstInt32 i = 0; i < ev->numEvents; ++i)
	{ 
		VstMidiEvent* event = (VstMidiEvent*)ev->events[i];

		if(event->type!=kVstMidiType) continue; 

		char* midiData=event->midiData;

		VstInt32 channel=midiData[0]&0x0f;
		VstInt32 status =midiData[0]&0xf0;

		switch(status)
		{
		case 0x80:	//note off
		case 0x90:	//note on
			{
				VstInt32 note=midiData[1]&0x7f; 

				VstInt32 velocity=midiData[2]&0x7f;

				if(status==0x80) velocity=0; 

				MidiAddNote(event->deltaFrames,channel,note,velocity);
			} 
			break;

		case 0xb0:	//control change
			{ 
				if(midiData[1]==0x64) MidiRPNLSB =midiData[2]&0x7f;
				if(midiData[1]==0x65) MidiRPNMSB =midiData[2]&0x7f;
				if(midiData[1]==0x26) MidiDataLSB=midiData[2]&0x7f;

				if(midiData[1]==0x06)
				{
					MidiDataMSB=midiData[2]&0x7f;

					if(MidiRPNLSB==0&&MidiRPNMSB==0) MidiPitchBendRange[channel]=(float)MidiDataMSB*.5f;
				}

				if (midiData[1] == 15 || midiData[1] == 20 || midiData[1] == 21 || midiData[1] == 22)//unused controllers 15,20,21,22
				{
					MidiAddControl(event->deltaFrames,channel,midiData[1]);//remember controller
				}

				if(midiData[1]>=0x7b)//all notes off and mono/poly mode changes that also requires to do all notes off
				{
					MidiAddNote(event->deltaFrames,channel,-1,0);
				}
			} 
			break;

		case 0xc0:	//program change
			{
				MidiAddProgramChange(event->deltaFrames,channel,midiData[1]&0x7f);
			}
			break;

		case 0xe0:	//pitch bend change
			{
				VstInt32 wheel = (midiData[1] & 0x7f) | ((midiData[2] & 0x7f) << 7);

				MidiAddPitchBend(event->deltaFrames,channel,(float)(wheel-0x2000)*MidiPitchBendRange[channel]/8192.0f);
			}
			break;
		}
	} 

	return 1; 
} 



VstInt32 PETCB2::SynthAllocateVoice(VstInt32 midi_ch,VstInt32 note)
{
	VstInt32 chn;

	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].note==note) return chn;
	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].note<0) return chn;

	return -1;
}



void PETCB2::SynthChannelChangeNote(VstInt32 chn,VstInt32 midi_ch,VstInt32 note)
{
	VstInt32 ins,key;

	key=note;
	ins=MidiProgram[midi_ch];

	if(MultiMode[midi_ch])
	{
		ins+=(note-60);

		while(ins<0) ins+=NUM_PROGRAMS;
		while(ins>=NUM_PROGRAMS) ins-=NUM_PROGRAMS;

		note=60;
	}

	SynthChannel[chn].note=note;
	SynthChannel[chn].key=key;
	SynthChannel[chn].midi_ch=midi_ch;

	SynthChannel[chn].ptr_vol = 0;
	SynthChannel[chn].ptr_arp = 0;
	SynthChannel[chn].ptr_pitch=0;
	SynthChannel[chn].ptr_wave = 0;

	SynthChannel[chn].pitch_acc = 0;
	SynthChannel[chn].arp_offset = 0;
	SynthChannel[chn].vol_output = 0;
	SynthChannel[chn].wave_output = 0;

	SynthChannel[chn].env_vol = &Instrument[ins].EnvVol;
	SynthChannel[chn].env_arp = &Instrument[ins].EnvArp;
	SynthChannel[chn].env_pitch = &Instrument[ins].EnvPitch;
	SynthChannel[chn].env_wave = &Instrument[ins].EnvWave;
}



void PETCB2::SynthChannelReleaseNote(VstInt32 midi_ch,VstInt32 note)
{
	int chn;

	if(note>=0)
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].key==note) SynthChannel[chn].note=-1;
		}
	}
	else
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch||midi_ch<0) SynthChannel[chn].note=-1;
		}
	}
}



void PETCB2::processReplacing(float **inputs,float **outputs,VstInt32 sampleFrames)
{
	float *outL=outputs[0];
	float *outR=outputs[1];
	float level,loops,freq,sample_rate;
	unsigned int i;
	int chn,add,vol_max;

	sample_rate=(float)updateSampleRate();

	while(--sampleFrames>=0)
	{
		for(i=0;i<MidiQueue.size();++i)
		{
			if(MidiQueue[i].delta>=0)
			{
				--MidiQueue[i].delta;

				if(MidiQueue[i].delta<0)//new message
				{
					switch(MidiQueue[i].type)
					{
					case mEventTypeNote:
						{
							if(MidiQueue[i].velocity)//key on
							{
								MidiKeyState[MidiQueue[i].note]=1;

								chn=SynthAllocateVoice(MidiQueue[i].channel,MidiQueue[i].note);

								if(chn>=0)
								{
									SynthChannelChangeNote(chn,MidiQueue[i].channel,MidiQueue[i].note);
								}
							}
							else//key off
							{
								if(MidiQueue[i].note>=0)
								{
									MidiKeyState[MidiQueue[i].note]=0;

									SynthChannelReleaseNote(MidiQueue[i].channel,MidiQueue[i].note);
								}
								else
								{
									memset(MidiKeyState,0,sizeof(MidiKeyState));

									SynthChannelReleaseNote(-1,-1);
								}
							}
						}
						break;

					case mEventTypePitchBend:
						{
							MidiPitchBend[MidiQueue[i].channel]=MidiQueue[i].depth;
						}
						break;

					case mEventTypeProgram:
						{
							if(MidiProgram[MidiQueue[i].channel]!=MidiQueue[i].program)
							{
								MidiProgram[MidiQueue[i].channel]=MidiQueue[i].program;

								SetUpdateGUI();
							}
						}
						break;

					case mEventTypeControl:
						{
							if(LogActive&&LogAuto)
							{
								if (MidiQueue[i].value == 15)
								{
									LogFileClose();
									++LogFilePart;
									LogFileOpen();
								}
								else if (MidiQueue[i].value == 20)
								{
									if (!LogAutoEnabled)
									{
										LogAutoEnabled = true;
									}
								}
								else if (MidiQueue[i].value == 21)
								{
									if(LogAutoEnabled)
									{
										LogAutoEnabled=false;

										LogStop();

										SetUpdateGUI();
									}
								}
							}

							if(MidiQueue[i].value==22)	//time sync marker
							{
								LogMarker=true;
							}
								
						}
						break;
					}
				}
			}
		}

		//advance envelopes and update counter

		sFrameAcc+=1.0f/(sample_rate/FrameRate);

		while (sFrameAcc >= 1.0f)
		{
			sFrameAcc -= 1.0f;

			for (chn = 0; chn < SYNTH_CHANNELS; ++chn)
			{
				if (SynthChannel[chn].note < 0) continue;

				if (SynthChannel[chn].ptr_vol >= SynthChannel[chn].env_vol->len)
				{
					if (!SynthChannel[chn].vol_output && !SynthChannel[chn].env_vol->has_loop) //terminate channel that is looped to zero volume at end
					{
						SynthChannel[chn].note = -1;
						continue;
					}

					SynthChannel[chn].ptr_vol = SynthChannel[chn].env_vol->loop_ptr;
				}

				SynthChannel[chn].vol_output = SynthChannel[chn].env_vol->value[SynthChannel[chn].ptr_vol++];

				if (SynthChannel[chn].ptr_wave >= SynthChannel[chn].env_wave->len) SynthChannel[chn].ptr_wave = SynthChannel[chn].env_wave->loop_ptr;

				SynthChannel[chn].wave_output = SynthChannel[chn].env_wave->value[SynthChannel[chn].ptr_wave++] & 15;

				if (SynthChannel[chn].ptr_arp >= SynthChannel[chn].env_arp->len) SynthChannel[chn].ptr_arp = SynthChannel[chn].env_arp->loop_ptr;

				if (!SynthChannel[chn].env_arp->abs[SynthChannel[chn].ptr_arp])
				{
					SynthChannel[chn].arp_offset = SynthChannel[chn].env_arp->value[SynthChannel[chn].ptr_arp++];
				}
				else
				{
					SynthChannel[chn].note = SynthChannel[chn].env_arp->value[SynthChannel[chn].ptr_arp++];
					SynthChannel[chn].arp_offset = 0;
				}

				if (SynthChannel[chn].ptr_pitch >= SynthChannel[chn].env_pitch->len) SynthChannel[chn].ptr_pitch = SynthChannel[chn].env_pitch->loop_ptr;

				SynthChannel[chn].pitch_acc += (float)SynthChannel[chn].env_pitch->value[SynthChannel[chn].ptr_pitch++];
			}

			//play channels

			sVIAWave = 0;
			sVIAReload = 0;

			vol_max = -1;

			for (i = 0; i < SYNTH_CHANNELS; ++i)
			{
				chn = SynthChannelOrder[i];

				if (SynthChannel[chn].note < 0) continue;

				if (SynthChannel[chn].vol_output == 0) continue;

				if (SynthChannel[chn].wave_output == 0) continue;

				if (SynthChannel[chn].vol_output < vol_max) continue;

				vol_max = SynthChannel[chn].vol_output;

				freq = (float)(SynthChannel[chn].note - 69);

				freq += MidiPitchBend[SynthChannel[chn].midi_ch];

				freq += SynthChannel[chn].arp_offset;

				freq += SynthChannel[chn].pitch_acc / 100.0f;

				sVIAReload = (int)(1000000.0f / 2 / 8 / (440.0f*pow(2.0f, freq / 12.0f)));
				sVIAWave = SynthChannel[chn].wave_output;
			}

			if (!RangeExpanded)
			{
				if (sVIAReload < 0 || sVIAReload>255)
				{
					sVIAReload = 0;
					sVIAWave = 0;
				}
			}
			else
			{
				if (sVIAReload < 0 || sVIAReload>65535)
				{
					sVIAReload = 0;
					sVIAWave = 0;
				}
			}
		
			if (sVIAWave != sVIAWavePrev)
			{
				sVIAWavePrev = sVIAWave;
				sVIAWaveShift = sVIAWave;
			}

			LogFrame();
		}

		//generate sound

		level=0;
		loops=0;
		add=(int)sample_rate;

		while (sVIAResampleAcc < 1000000 / 2)
		{
			if (sVIAOutput) level += 1.0f;

			if (sVIAReload > 0)
			{
				--sVIACount;

				if (sVIACount < 0)
				{
					sVIACount += sVIAReload;
					sVIAOutput = sVIAWaveShift & 1;
					if (sVIAWaveShift & 1) sVIAWaveShift |= 0x100;
					sVIAWaveShift = sVIAWaveShift >> 1;
				}
			}

			sVIAResampleAcc+=add;

			loops+=1.0f;
		}

		sVIAResampleAcc -= 1000000 / 2;

		level/=loops;

		level*=pOutputGain;

		(*outL++)=level;
		(*outR++)=level;
	}

	MidiQueue.clear();

	UpdateGUI();
}
