#include <windows.h>
#include <math.h>
#include <vector>

using namespace std;

#include "audioeffectx.h"



#define PLUGIN_NAME				"PETCB2"
#define PLUGIN_VENDOR			"Shiru"
#define PLUGIN_PRODUCT			"v1.1 20.10.22"
#define PLUGIN_UID				'pcb2'

#define NUM_INPUTS				0
#define NUM_OUTPUTS				2
#define NUM_PROGRAMS			128
#define NUM_ENVELOPES			4

#define MIDI_CHANNELS_ALL		16

#define CHUNK_SIZE				(NUM_PROGRAMS*NUM_ENVELOPES*ENVELOPE_TEXT_LEN*2)

#define MAX_NAME_LEN			32

#define SYNTH_CHANNELS			128

#define ENVELOPE_TEXT_LEN		1024
#define ENVELOPE_DATA_LEN		2048
#define ENVELOPE_NUM_RANGE		0x8000



enum {

	NUM_PARAMS=0

};



enum {

	mEventTypeNote=0,
	mEventTypePitchBend,
	mEventTypeProgram,
	mEventTypeControl

};



struct MidiQueueStruct
{
	VstInt32 type;
	VstInt32 delta;
	VstInt32 channel;
	VstInt32 note;
	VstInt32 velocity;
	VstInt32 program;
	VstInt32 value;

	float depth;
};



enum {

	PARSE_PREFIX_NONE=0,
	PARSE_PREFIX_NEGATIVE,
	PARSE_PREFIX_REPEAT,
	PARSE_PREFIX_TIME_SCALE,
	PARSE_PREFIX_ABSOLUTE

};



enum {

	LOG_FORMAT_ASM=1,
	LOG_FORMAT_H,
	LOG_FORMAT_BIN

};

#define LOG_MAX_FRAMES	254
#define LOG_MARKER_TAG	255



struct envelopeStruct {

	VstInt32 value[ENVELOPE_DATA_LEN];
	VstInt8 abs[ENVELOPE_DATA_LEN];

	VstInt32 len;
	VstInt32 loop_ptr;

	bool has_loop;
	bool error;

};



struct instrumentStruct {

	envelopeStruct EnvVol;
	envelopeStruct EnvArp;
	envelopeStruct EnvPitch;
	envelopeStruct EnvWave;

};



struct channelStruct {

	VstInt32 midi_ch;
	VstInt32 note;
	VstInt32 key;
	
	envelopeStruct *env_vol;
	envelopeStruct *env_arp;
	envelopeStruct *env_pitch;
	envelopeStruct *env_wave;

	VstInt32 ptr_vol;
	VstInt32 ptr_arp;
	VstInt32 ptr_pitch;
	VstInt32 ptr_wave;

	VstInt32 arp_offset;
	VstInt32 vol_output;
	VstInt32 wave_output;

	float pitch_acc;

};




#define TAG(s)		(float)(((s)[0])|(((s)[1])<<7)|(((s)[2])<<14)|(((s)[3]<<21)))

#define DATA		TAG("DATA")
#define MPRG		TAG("MPRG")
#define PROG		TAG("PROG")
#define NAME		TAG("NAME")
#define EVOL		TAG("EVOL")
#define EARP		TAG("EARP")
#define EPTC		TAG("EPTC")
#define EWAV		TAG("EWAV")
#define FRMR		TAG("FRMR")
#define MMOD		TAG("MMOD")
#define GAIN		TAG("GAIN")
#define DONE		TAG("DONE")



const char* const PresetData[]={
	"Default","4","0","0","15",
	"Bass","2","0","0","15",
	"Solo","6","12 0","0*16 x4 | 8 -8","15",
	"Kick","8*3 0","#64","| -800","15",
	"Snare","8 8 8 0 8 0","#64 -12 0","| -240","15",
	"Short arp","4 4 4 0","12 0","0","15",
	"Truu","4 0 4 0 4 0 4 0","0","| -100","15",
	""
};



class PETCB2:public AudioEffectX
{
public:

	PETCB2(audioMasterCallback audioMaster);

	~PETCB2();

	virtual void setParameter(VstInt32 index,float value);
	virtual float getParameter(VstInt32 index);
	virtual void getParameterLabel(VstInt32 index,char *label);
	virtual void getParameterDisplay(VstInt32 index,char *text);
	virtual void getParameterName(VstInt32 index,char *text); 

	virtual VstInt32 getProgram();
	virtual void setProgram(VstInt32 program);
	virtual void getProgramName(char* name);
	virtual void setProgramName(char* name);

	virtual bool getEffectName(char *name) { strcpy(name,PLUGIN_NAME); return true; }
	virtual bool getVendorString(char *text) { strcpy(text,PLUGIN_VENDOR); return true; }
	virtual bool getProductString(char *text) { strcpy(text, PLUGIN_PRODUCT); return true; }
	virtual VstInt32 getVendorVersion() { return 1000; } 

	VstInt32 canDo(char* text);
	VstInt32 getNumMidiInputChannels(void);
	VstInt32 getNumMidiOutputChannels(void);

	virtual VstInt32 getChunk(void** data,bool isPreset=false);
	virtual VstInt32 setChunk(void* data,VstInt32 byteSize,bool isPreset=false);

	VstInt32 processEvents(VstEvents* ev);
	void processReplacing(float**inputs,float **outputs,VstInt32 sampleFrames);

	VstInt32 MidiChannel;
	VstInt32 MidiProgram[MIDI_CHANNELS_ALL];
	VstInt32 MidiProgramPrev[MIDI_CHANNELS_ALL];

	char ProgramName[NUM_PROGRAMS][MAX_NAME_LEN];

	char EnvelopeV[NUM_PROGRAMS][ENVELOPE_TEXT_LEN];
	char EnvelopeA[NUM_PROGRAMS][ENVELOPE_TEXT_LEN];
	char EnvelopeP[NUM_PROGRAMS][ENVELOPE_TEXT_LEN];
	char EnvelopeW[NUM_PROGRAMS][ENVELOPE_TEXT_LEN];

	const char* CompileEnvelope(char *str,envelopeStruct *env,bool hold);
	void CompileInstruments(void);

	instrumentStruct Instrument[NUM_PROGRAMS];

	void SetUpdateGUI(void);
	void UpdateGUI(void);

	float FrameRate;
	bool MultiMode[16];

	bool LogActive;
	VstInt32 LogFilePart;
	VstInt32 LogFrameDelta;
	VstInt32 LogDiv;
	VstInt32 LogDivPrev;
	VstInt32 LogWave;
	VstInt32 LogWavePrev;
	char LogFileName[MAX_PATH];
	FILE *LogFile;
	VstInt32 LogFormat;
	bool RangeExpanded;
	bool LogAuto;
	bool LogAutoEnabled;
	bool LogMarker;

	void LogStart(void);
	void LogStop(void);
	void LogFrameFlush(void);
	void LogFrame(void);
	void LogFileOpen(void);
	void LogFileClose(void);

protected:

	float pOutputGain;

	VstInt32 SavePresetChunk(float *chunk);
	void LoadPresetChunk(float *chunk);

	VstInt32 SaveStringChunk(char *str,float *dst);
	VstInt32 LoadStringChunk(char *str,int max_length,float *src);

	float Chunk[CHUNK_SIZE];

	void MidiAddNote(VstInt32 delta,VstInt32 channel,VstInt32 note,VstInt32 velocity);
	void MidiAddPitchBend(VstInt32 delta,VstInt32 channel,float depth);
	void MidiAddProgramChange(VstInt32 delta,VstInt32 channel,VstInt32 program);
	void MidiAddControl(VstInt32 delta,VstInt32 channel,VstInt32 value);

	bool MidiIsAnyKeyDown(void);

	vector<MidiQueueStruct> MidiQueue;

	unsigned char MidiKeyState[128];

	float MidiPitchBend[16];
	float MidiPitchBendRange[16];

	VstInt32 MidiRPNLSB;
	VstInt32 MidiRPNMSB;
	VstInt32 MidiDataLSB;
	VstInt32 MidiDataMSB;

	float sFrameAcc;
	
	VstInt32 sVIAOutput;
	VstInt32 sVIACount;
	VstInt32 sVIAReload;
	VstInt32 sVIAWave;
	VstInt32 sVIAWavePrev;
	VstInt32 sVIAWaveShift;
	VstInt32 sVIAResampleAcc;

	channelStruct SynthChannel[SYNTH_CHANNELS];

	VstInt32 SynthChannelOrder[SYNTH_CHANNELS];

	VstInt32 SynthAllocateVoice(VstInt32 midi_ch,VstInt32 note);
	void SynthChannelChangeNote(VstInt32 chn,VstInt32 midi_ch,VstInt32 note);
	void SynthChannelReleaseNote(VstInt32 midi_ch,VstInt32 note);

	bool UpdateGuiFlag;
};
